#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904

from hamcrest import is_
from hamcrest import is_not
from hamcrest import raises
from hamcrest import calling
from hamcrest import assert_that
from hamcrest import same_instance
does_not = is_not

import fudge

from zope.interface import implementer

from zope.component import globalSiteManager as BASE

from zope.component.hooks import getSite
from zope.component.hooks import setSite

from zope.site.site import LocalSiteManager as LSM

from z3c.baseregistry.baseregistry import BaseComponents

from zope.site.interfaces import IRootFolder

from nti.site.interfaces import IMainApplicationFolder

from nti.site.subscribers import threadSiteSubscriber

from nti.site.transient import HostSiteManager as HSM
from nti.site.transient import TrivialSite

from nti.testing.base import AbstractTestBase


class TestSubscriber(AbstractTestBase):

    def _check_site(self, new_site, expected_after_subscriber):
        threadSiteSubscriber(new_site, None)
        assert_that(getSite(), expected_after_subscriber)

    def test_root_folder_not_set_if_site_installed(self):
        installed = TrivialSite(LSM(None))
        setSite(installed)

        @implementer(IRootFolder)
        class Root(object):
            pass

        self._check_site(Root(), is_(same_instance(installed)))

    def test_main_folder_not_set_if_site_installed(self):
        installed = TrivialSite(LSM(None))
        setSite(installed)

        @implementer(IMainApplicationFolder)
        class Main(object):
            pass

        self._check_site(Main(), is_(same_instance(installed)))

    def test_root_folder_set_if_no_site_installed(self):
        @implementer(IRootFolder)
        class Root(object):
            def getSiteManager(self):
                return LSM(None)

        self._check_site(Root(), is_(Root))

    def test_main_folder_set_if_no_site_installed(self):
        @implementer(IMainApplicationFolder)
        class Main(object):
            def getSiteManager(self):
                return LSM(None)

        self._check_site(Main(), is_(Main))

    @fudge.patch('nti.site.subscribers.getSite')
    def test_not_replace_same(self, fake_get):
        fake_get.is_callable().returns(self)

        # If we actually tried to set TestSubscriber as the site,
        # we'd blow up
        threadSiteSubscriber(self, None)

    @fudge.patch('nti.site.subscribers.getSite')
    def test_returns_unsettable_proxy(self, fake_get):
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))
        host_comps = BaseComponents(BASE, 'example.com', (BASE,))
        host_sm = HSM('example.com', 'siteman', host_comps, pers_comps)

        fake_get.is_callable().returns(TrivialSite(host_sm))

        host_comps2 = BaseComponents(BASE, 'other.com', (BASE,))
        host_sm2 = HSM('other.com', 'siteman', host_comps2, pers_comps)

        threadSiteSubscriber(TrivialSite(host_sm2), None)

        # Existing host comps are preserved
        assert_that(getSite().getSiteManager().host_components, is_(host_comps))

        # and it's proxied and we can't change it
        assert_that(calling(getSite().setSiteManager).with_args(None),
                    raises(ValueError, "Cannot set site manager on proxy"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Support for testing code that uses :mod:`nti.site`.

Most of the functionality exposed through this module uses :mod:`ZODB`
to test persistence and transaction handling and is based on the
support code from :mod:`nti.testing.zodb`.

.. versionadded:: 2.1.0
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division


import unittest
import functools

from ZODB.DemoStorage import DemoStorage
from hamcrest import assert_that

from nti.testing import zodb
from nti.testing.layers import ConfiguringLayerMixin
from nti.testing.layers import GCLayerMixin
from nti.testing.layers import find_test
from nti.testing.matchers import provides

from zope import component
from zope.component.hooks import setHooks
from zope.component.hooks import resetHooks
from zope.component.hooks import site as currentSite
from zope.site import SiteManagerContainer
import ZODB
import zope.testing.cleanup

from .hostpolicy import install_main_application_and_sites
from . import hostpolicy
from .interfaces import IMainApplicationFolder
from .site import BTreeLocalSiteManager
from .site import get_site_for_site_names

__all__ = [
    'setHooks',
    'resetHooks',
    'print_tree',
    'persistent_site_trans',
    'SharedConfiguringTestLayer',
    'SiteTestCase',
    'uses_independent_db_site',
]

def format_tree(folder, **kwargs):
    """
    Like :func:`print_tree`, but returns the results as a string
    instead of printing them to ``stdout`` or a file.

    Any *file* argument given is ignored.

    .. versionadded:: 2.2.0
    """
    if str is bytes:
        # Handles unicode and bytes mixed appropriately
        import cStringIO as io
    else:
        import io
    out = io.StringIO()
    kwargs['file'] = out
    print_tree(folder, **kwargs)
    return out.getvalue()


def print_tree(folder, **kwargs):
    """
    print_tree(folder, file=sys.stdout, show_unknown=repr, basic_indent='    ', details=('id', 'type', 'len', 'siteManager')) -> None

    Print a descriptive tree of the contents of the dict-like *folder* to *file*.

    Pass a subset of *details* to disable printing certain information
    when it isn't relavent.

    .. versionchanged:: 2.2.0
       Add several arguments including *show_unknown*, *details*.
       Print the contents of site managers by default.
       Fix a bug not passing the *basic_indent* to recursive calls.
    """
    # pylint:disable=too-many-locals,too-many-branches,too-many-statements
    # XXX: Refactor me.
    import sys
    from zope.site.interfaces import IRootFolder
    from zope.component.interfaces import ISite

    file = kwargs.get('file', sys.stdout)
    seen = kwargs.get('seen')
    if seen is None:
        seen = kwargs['seen'] = dict()
    basic_indent = kwargs.get('basic_indent', '    ')
    depth = kwargs.get('depth', 1)
    details = kwargs.get('details', ('id', 'type', 'len', 'siteManager'))
    name = kwargs.get('name', None)
    show_unknown = kwargs.get('show_unknown', repr)
    known_types = kwargs.get('known_types', (int, str, type(u''), float, type(None)))
    extra_details = kwargs.get('extra_details', lambda o: ())

    indent = basic_indent * depth
    folder_id = id(folder)

    if name is None:
        name = getattr(folder, '__name__', None)
        if not name:
            if getattr(folder, '_p_jar', None) and folder._p_jar.root() is folder:
                name = "<Connection Root Dictionary>"
            else:
                name = str(folder)

    if folder_id in seen:
        print(indent, name, '->', seen[folder_id], file=file)
        return
    seen[folder_id] = name + ' ' + str(folder_id)

    provs = []
    for iface in ISite, IRootFolder, IMainApplicationFolder:
        if iface.providedBy(folder):
            provs.append(iface.__name__)
    if provs:
        name = '<' + ','.join(provs) + '>: ' + name

    print_args = []
    if indent:
        print_args.append(indent)
    print_args.append(name)
    if 'id' in details:
        print_args.append(folder_id)
    if 'len' in details:
        try:
            print_args.append('len=' + str(len(folder)))
        except TypeError:
            pass
    if 'type' in details:
        print_args.append(type(folder))

    print_args.extend(extra_details(folder))

    if not hasattr(folder, 'items'):
        print_args.append('=>')
        if isinstance(folder, known_types):
            print_args.append(repr(folder))
        else:
            print_args.append(show_unknown(folder))

    print(*print_args, file=file)

    if hasattr(folder, 'items'):
        recur_args = kwargs.copy()
        recur_args['depth'] = depth + 1
        for k, v in sorted(folder.items()):
            recur_args['name'] = k
            print_tree(v, **recur_args)

    if 'siteManager' in details and ISite.providedBy(folder):
        site_man = folder.getSiteManager()
        recur_args = kwargs.copy()
        recur_args['depth'] = depth + 1
        recur_args['name'] = '<Site Manager> name=' + site_man.__name__
        print_tree(site_man, **recur_args)


class persistent_site_trans(zodb.mock_db_trans):
    """
    Context manager for a ZODB database connection and
    active ``zope.component`` site (usually) persisted in the database.

    .. versionchanged:: 2.1.0
       While there was no previous public version of this class,
       there was a private version in ``nti.site.tests``. That version
       called :func:`.install_main_application_and_sites` setting
       the *root_alias* to the :attr:`main_application_folder_name`.
       Since older versions of that function installed the *root_alias* to point
       to the *main_name*, this used result in no alias for the root folder in the
       root. Now, there will be an alias (at :obj:`~.DEFAULT_ROOT_ALIAS`).
    """

    #: The site to make active by default and when looking up
    #: a *site_name*. This must identify an object in the root
    #: of the database that provides :class:`~.IMainApplicationFolder`.
    main_application_folder_name = hostpolicy.DEFAULT_MAIN_ALIAS

    def __init__(self, db=None, site_name=None):
        """
        :param db: See :class:`nti.testing.zodb.mock_db_trans`
        :keyword str site_name: The name of a site to be made current
            during execution of the body.
            The site is found using :func:`~.get_site_for_site_names`
            while :attr:`main_application_folder_name` is the current site.
            If not given, the site found at :attr:`main_application_folder_name`
            will be the current site.
        """
        super(persistent_site_trans, self).__init__(db)
        self._site_cm = None
        self._site_name = site_name

    def on_application_and_sites_installed(self, folder):
        """
        Called when the main application and sites have been installed. This
        may not be called every time an instance of this class is used, as
        the database may be persistent.
        """
        assert_that(folder, provides(IMainApplicationFolder))

    def on_main_application_folder_missing(self, conn):
        """
        Called from :meth:`on_connection_opened` when the :attr:`main_application_folder_name`
        is not found in the root of the *conn*.

        This method calls :func:`~.install_main_application_and_sites`, passing
        :attr:`main_application_folder_name` as the *main_alias*.
        """
        install_main_application_and_sites(conn,
                                           main_alias=self.main_application_folder_name,
                                           main_setup=self.on_application_and_sites_installed)

    def on_connection_opened(self, conn):
        super(persistent_site_trans, self).on_connection_opened(conn)
        main_name = hostpolicy.text_type(self.main_application_folder_name)

        root = conn.root()
        if main_name not in root:
            self.on_main_application_folder_missing(conn)

        sitemanc = conn.root()[hostpolicy.text_type(self.main_application_folder_name)]
        if self._site_name:
            with currentSite(sitemanc):
                sitemanc = get_site_for_site_names((self._site_name,))

        self._site_cm = currentSite(sitemanc)
        self._site_cm.__enter__() # pylint:disable=no-member
        assert component.getSiteManager() == sitemanc.getSiteManager()
        return conn

    def __exit__(self, t, v, tb):
        result = self._site_cm.__exit__(t, v, tb) # pylint:disable=no-member
        super(persistent_site_trans, self).__exit__(t, v, tb)
        return result


mock_db_trans = persistent_site_trans # BWC, remove in 2021
reset_db_caches = zodb.reset_db_caches # BWC, remove in 2021

def _mock_ds_wrapper_for(func, installer_factory, installer_kwargs, db_factory,
                         marker=object()):

    @functools.wraps(func)
    def f(*args):
        # we may not be in a layer that's done this. Note that we don't tear it down though.
        # This is needed to run ``install_main_application_and_sites``
        setHooks()
        db = db_factory()

        old_db = zodb.ZODBLayer.db
        old_func_db = marker
        try:
            zodb.ZODBLayer.db = db
            if SharedConfiguringTestLayer.current_test is not None:
                SharedConfiguringTestLayer.current_test.db = db
            if args and isinstance(args[0], unittest.TestCase):
                # self
                old_func_db = getattr(args[0], 'db', marker)
                args[0].db = db

            with installer_factory(db, **installer_kwargs):
                pass

            sitemanc = SiteManagerContainer()
            sitemanc.setSiteManager(BTreeLocalSiteManager(None))

            with currentSite(sitemanc):
                assert component.getSiteManager() == sitemanc.getSiteManager()
                func(*args)
        finally:
            if args and isinstance(args[0], unittest.TestCase):
                if old_func_db is marker:
                    del args[0].db
                else:
                    args[0].db = old_func_db
            db.close()
            zodb.ZODBLayer.db = old_db
    return f


def default_db_factory():
    return ZODB.DB(DemoStorage(name='Users'))

def uses_independent_db_site(*args, **kwargs):
    """
    uses_independent_db_site(db_factory=None, installer_factory=persistent_site_trans, installer_kwargs={}) -> function

    A decorator or decorator factory. Creates a new database using *db_factory*,
    initializes it using *installer_factory*, and then runs the body of the function
    in a site and site manager that are disconnected from the database.

    If the function is a unittest method, the unittest object's ``db`` attribute
    will be set to the created db during executing. Likewise, the :class:`nti.testing.zodb.ZODBLayer`
    ``db`` attribute (and layers that extend from it like :class:`SharedConfiguringTestLayer`)
    will be set to this object and returned to the previous value on exit.

    This can be called as given in the signature, or can be called with no arguments::

        class MyTest(TestCase):

            @uses_independent_db_site
            def test_something(self):
                pass

            @uses_independent_db_site(installer_factory=MyCustomFactory)
            def test_something_else(self):
                pass

    The body of the function is free to use :class:`persistent_site_trans` statements.
    They will default to using the database established here instead of
    a database established by a test layer (which should be the same in most cases).

    :keyword callable db_factory: The 0-argument factory used to create a DB to pass
       to the installer.

    :keyword type installer_factory: The factory used to create
       the installer. The installer executes before the body of the wrapped
       function. This defaults to :class:`persistent_site_trans`, but can be
       set to any custom subclass that accepts the db as its first argument and
       is a context manager that does whatever installation is needed and commits
       the transaction.

    :keyword dict installer_kwargs: Keyword arguments to pass to the *installer_factory*
    """
    assert bool(args) ^ bool(kwargs), "Cannot mix keyword arguments and regular arguments."

    installer_factory = kwargs.pop('installer_factory', persistent_site_trans)
    installer_kwargs = kwargs.pop('installer_kwargs', {})
    db_factory = kwargs.pop('db_factory', default_db_factory)
    assert not kwargs
    assert len(args) == 1 or not args

    func_factory = lambda func: _mock_ds_wrapper_for(func,
                                                     installer_factory,
                                                     installer_kwargs,
                                                     db_factory)
    if len(args) == 1:
        # Being used as a plain decorator
        return func_factory(args[0])
    return func_factory


class SharedConfiguringTestLayer(zodb.ZODBLayer,
                                 GCLayerMixin,
                                 ConfiguringLayerMixin):
    """
    A test layer that configures this package, and sets other useful
    test settings.

    Note that the details of the test settings may change. In this version,
    this configures the :class:`~.BTreeLocalAdapterRegistry` to immediately
    switch to BTrees instead of dicts.
    """

    set_up_packages = ('nti.site',)

    #: The test (a :class:`unittest.TestCase` subclass) currently
    #: executing in this layer. If there is no such test, this is `None`.
    current_test = None

    @classmethod
    def setUp(cls):
        setHooks()
        cls.setUpPackages()
        # Force all the thresholds low so that we do as much testing as possible
        # with btrees.
        from .folder import HostPolicySiteManager
        assert hasattr(HostPolicySiteManager, 'btree_threshold')
        HostPolicySiteManager.btree_threshold = 0

    @classmethod
    def tearDown(cls):
        from .site import BTreeLocalAdapterRegistry
        from .folder import HostPolicySiteManager
        del HostPolicySiteManager.btree_threshold
        assert hasattr(HostPolicySiteManager, 'btree_threshold')
        cls.tearDownPackages()
        zope.testing.cleanup.cleanUp()

    @classmethod
    def testSetUp(cls, test=None): # pylint:disable=arguments-differ
        """
        Tests that run in this layer have their ``db`` property set to the
        current ``db`` of this layer.
        """
        setHooks()
        cls.current_test = test = test or find_test()
        test.db = cls.db

    @classmethod
    def testTearDown(cls):
        """
        When a test in this layer is torn down, its ``db`` property is set
        to ``None``, as is this layer's :attr:`current_test`.
        """
        cls.current_test.db = None
        cls.current_test = None



class SiteTestCase(unittest.TestCase):
    """
    A test case that runs in the :class:`SharedConfiguringTestLayer`.
    """
    layer = SharedConfiguringTestLayer



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904


from hamcrest import assert_that
from hamcrest import is_


from nti.testing import base

import six
import transaction

from zope import component
from ZODB.interfaces import IDatabase

import ZODB.DB
from ZODB.DemoStorage import DemoStorage
from zope.site import SiteManagerContainer


from ..runner import run_job_in_site
from ..runner import _tx_string

from ..transient import TrivialSite

from ..interfaces import SiteNotInstalledError


class TestRunner(base.AbstractTestBase):

    def setUp(self):
        super(TestRunner, self).setUp()
        db = ZODB.DB(DemoStorage(name='base'))
        component.provideUtility(db, IDatabase)

        conn = db.open()
        smc = conn.root()[u'nti.dataserver'] = SiteManagerContainer()
        smc.setSiteManager(component.getGlobalSiteManager())

        transaction.commit()
        conn.close()

    def test_site_cm_success(self):
        class MockConn(object):
            def __init__(self):
                self._root = {}
            def root(self):
                return self._root

        c = MockConn()
        c.root()[u'nti.dataserver'] = TrivialSite(component.getGlobalSiteManager())
        def handler():
            return 42

        result = run_job_in_site(handler)
        self.assertEqual(result, 42)

    def test_site_cm_not_installed(self):
        class MockConn(object):
            def __init__(self):
                self._root = {}
            def root(self):
                return self._root
            closed = False
            def close(self):
                self.closed = True

        c = MockConn()
        c._root[u'nti.dataserver'] = TrivialSite(None)

        class MockDB(object):
            def open(self):
                return c

        component.provideUtility(MockDB(), IDatabase)

        with self.assertRaises(SiteNotInstalledError):
            run_job_in_site(lambda: None)

        self.assertTrue(c.closed)

    def test_run_description(self):
        expected_desc = None
        expected_desc_type = six.text_type

        # Note: This file's coding is utf-8!
        def func():
            "A docstring with utf-8 chars: 😀"
            assert_that(component.getSiteManager(), is_(component.getGlobalSiteManager()))
            assert_that(transaction.get().description, is_(expected_desc))
            assert_that(transaction.get().description, is_(expected_desc_type))

        expected_desc = _tx_string('func\n\n' + func.__doc__)
        run_job_in_site(func, site_names=('abc',))

        expected_desc = _tx_string('Test')
        run_job_in_site(func, job_name=b"Test")

        func.__name__ = '_'
        expected_desc = _tx_string(func.__doc__)
        run_job_in_site(func, site_names=('abc',))

        expected_desc = None
        run_job_in_site(lambda: None)

    def test_connection_opened_in_explicit_mode(self):
        from nti.site.runner import _RunJobInSite
        test = self
        class Loop(_RunJobInSite):
            def run_handler(self, *args, **kwargs):
                test.assertTrue(self._connection.explicit_transactions)
                return 42

        result = Loop(lambda: None, site_names=(), job_name='', side_effect_free=False,
                      root_folder_name='nti.dataserver')()
        self.assertEqual(result, 42)

    def test_run_missing_name_doc(self):
        # Issue 16
        class Callable(object):
            # Like functools.partial, this doesn't expose a __name__ or __doc__.
            def __getattribute__(self, name):
                if name in ('__doc__', '__name__'):
                    raise AttributeError(name)
                return object.__getattribute__(self, name) # pragma: no cover

            def __call__(self):
                assert_that(transaction.get().description, is_(''))


        run_job_in_site(Callable())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904

from hamcrest import is_
from hamcrest import none
from hamcrest import not_none
from hamcrest import assert_that

from zope import component

from nti.site.interfaces import ISiteMapping

from nti.testing.base import ConfiguringTestBase

ZCML_STRING = """
<configure  xmlns="http://namespaces.zope.org/zope"
            xmlns:i18n="http://namespaces.zope.org/i18n"
            xmlns:zcml="http://namespaces.zope.org/zcml"
            xmlns:sites="http://nextthought.com/sites">

    <include package="zope.component" file="meta.zcml" />
    <include package="zope.security" file="meta.zcml" />
    <include package="zope.component" />
    <include package="." file="meta.zcml" />

    <configure>
        <sites:registerSiteMapping source_site_name="mySite1"
                                   target_site_name="mySite2" />
    </configure>
</configure>

"""


class TestZcml(ConfiguringTestBase):

    def test_registration(self):
        self.configure_string(ZCML_STRING)
        site_mapping = component.queryUtility(ISiteMapping, name='mySite1')
        assert_that(site_mapping, not_none())
        assert_that(site_mapping.source_site_name, is_('mySite1'))
        assert_that(site_mapping.target_site_name, is_('mySite2'))

        site_mapping = component.queryUtility(ISiteMapping, name='mySite2')
        assert_that(site_mapping, none())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904
# pylint:disable=too-many-ancestors

from hamcrest import is_
from hamcrest import is_not
from hamcrest import raises
from hamcrest import calling
from hamcrest import contains
from hamcrest import assert_that
from hamcrest import has_property
from hamcrest import same_instance
from hamcrest import none
from hamcrest import not_none
from hamcrest import has_length
does_not = is_not

import unittest

from zope import interface

from zope.interface import ro
from zope.interface import implementedBy
from zope.interface import Interface
from zope.interface.interfaces import ComponentLookupError

from zope.component.hooks import getSite
from zope.component.hooks import setSite
from zope.component.hooks import site as currentSite

from zope.interface.interfaces import IComponents


from zope.component import globalSiteManager as BASE

from z3c.baseregistry.baseregistry import BaseComponents

from nti.site.folder import HostSitesFolder

from nti.site.interfaces import IHostPolicyFolder
from nti.site.subscribers import threadSiteSubscriber
from nti.site.transient import HostSiteManager as HSM
from nti.site.transient import TrivialSite
from nti.site.site import get_site_for_site_names
from nti.site.site import find_site_components
from nti.site.site import get_component_hierarchy
from nti.site.site import get_component_hierarchy_names

from nti.site.tests import SharedConfiguringTestLayer

from nti.testing.matchers import validly_provides
from nti.testing.matchers import is_false
from nti.testing.base import AbstractTestBase

from persistent import Persistent

import fudge

class IMock(Interface): # pylint:disable=inherit-non-class
    pass

@interface.implementer(IMock)
class MockSite(object):

    __name__ = None
    __parent__ = None

    def __init__(self, site_man=None):
        self.site_man = site_man

    def getSiteManager(self):
        return self.site_man

class IFoo(Interface): # pylint:disable=inherit-non-class
    pass

@interface.implementer(IFoo)
class RootFoo(object):
    pass

class TestSiteSubscriber(unittest.TestCase):

    layer = SharedConfiguringTestLayer

    def testProxyHostComps(self):
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))
        host_comps = BaseComponents(BASE, 'example.com', (BASE,))
        host_sm = HSM('example.com', 'siteman', host_comps, pers_comps)
        host_site = MockSite(host_sm)
        host_site.__name__ = host_sm.__name__
        setSite(host_site)

        new_comps = BaseComponents(BASE, 'sub_site', (pers_comps,))
        new_site = MockSite(new_comps)
        new_site.__name__ = new_comps.__name__
        interface.alsoProvides(new_site, IFoo)

        threadSiteSubscriber(new_site, None)

        cur_site = getSite()
        # It should implement the static and dynamic
        # ifaces
        assert_that(cur_site, validly_provides(IFoo))
        assert_that(cur_site, validly_provides(IMock))

        # It should have the marker property
        assert_that(cur_site.getSiteManager(),
                    has_property('host_components',
                                 host_comps))

        assert_that(ro.ro(cur_site.getSiteManager()),
                    contains(
                        # The first entry is synthesized
                        has_property('__name__', new_comps.__name__),
                        pers_comps,
                        # The host comps appear after all the bases
                        # in the ro of the new site
                        host_comps,
                        BASE))

    def testTraverseFailsIntoSiblingSiteExceptHostPolicyFolders(self):
        new_comps = BaseComponents(BASE, 'sub_site', ())
        new_site = MockSite(new_comps)
        new_site.__name__ = new_comps.__name__

        with currentSite(None):
            threadSiteSubscriber(new_site, None)
            # If we walk into a site...
            # ...and then try to walk into a sibling site with no apparent relationship...
            new_comps2 = BaseComponents(BASE, 'sub_site', (new_comps,))
            new_site2 = MockSite(new_comps2)
            new_site2.__name__ = new_comps2.__name__

            # ... and that's allowed, completely replacing the current site...
            threadSiteSubscriber(new_site2, None)
            assert_that(getSite(), is_(same_instance(new_site2)))

            # ... reset ...
            threadSiteSubscriber(new_site, None)
            assert_that(getSite(), is_(same_instance(new_site)))

            # ... if they are both HostPolicyFolders...
            interface.alsoProvides(new_site, IHostPolicyFolder)
            interface.alsoProvides(new_site2, IHostPolicyFolder)
            threadSiteSubscriber(new_site2, None)

            # ... traversal does not change the site
            assert_that(getSite(), is_(same_instance(new_site)))

    def test_components_unregister_on_site_removal(self):
        site_folder = HostSitesFolder()
        new_site = MockSite()
        interface.alsoProvides(new_site, IHostPolicyFolder)
        key = 'new_site_name'
        site_folder[key] = new_site
        new_site_components = BaseComponents(BASE, 'new_site_name', (BASE,))
        current_site_manager = getSite().getSiteManager()
        current_site_manager.registerUtility(new_site_components,
                                             name=key,
                                             provided=IComponents)

        assert_that(current_site_manager.queryUtility(IComponents, name=key),
                    not_none())
        del site_folder[key]
        assert_that(current_site_manager.queryUtility(IComponents, name=key),
                    none())

        # Safe without registered components
        site_folder[key] = new_site
        del site_folder[key]


class TestGetSiteForSiteNames(AbstractTestBase):

    @fudge.patch('nti.site.site.find_site_components')
    def test_no_persistent_site(self, fake_find):
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))
        host_comps = BaseComponents(BASE, 'example.com', (BASE,))
        host_sm = HSM('example.com', 'siteman', host_comps, pers_comps)
        class PersistentTrivialSite(Persistent, TrivialSite):
            pass
        trivial_site = PersistentTrivialSite(host_sm)
        fake_find.is_callable().returns(pers_comps)

        from zope.interface.ro import C3
        C3.STRICT_IRO = False
        try:
            x = get_site_for_site_names(('',), trivial_site)
        finally:
            C3.STRICT_IRO = C3.ORIG_STRICT_IRO # pylint:disable=no-member

        assert_that(x, is_not(Persistent))
        assert_that(x, is_(TrivialSite))
        assert_that(x.__name__, is_(pers_comps.__name__))

    def test_find_comps_empty(self):
        assert_that(find_site_components(('',)),
                    is_(none()))


class TestGetComponentHierarchy(AbstractTestBase):

    @fudge.patch('nti.site.site.find_site_components')
    def test_get_component_hierarchy(self, fake_find):
        # This is a lousy test, just testing what we already see done, not
        # correct behaviour
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))

        class LocalMockSite(object):
            __parent__ = None
            __name__ = None

            def __init__(self):
                self.__parent__ = {}

        site_comps_1 = BaseComponents(pers_comps, '1', (pers_comps,))
        site_comps_2 = BaseComponents(site_comps_1, '2', (site_comps_1,))

        site = LocalMockSite()
        site.__parent__[site_comps_1.__name__] = site_comps_1
        site.__parent__[site_comps_2.__name__] = site_comps_2

        fake_find.is_callable().returns(site_comps_2)

        x = list(get_component_hierarchy(site))
        assert_that(x, is_([site_comps_2, site_comps_1]))

        x = list(get_component_hierarchy_names(site))
        assert_that(x, is_(['2', '1']))

        x = list(get_component_hierarchy_names(site, reverse=True))
        assert_that(x, is_(['1', '2']))

from nti.site.site import BTreeLocalSiteManager as BLSM
from nti.site.site import _LocalAdapterRegistry
from nti.site.site import BTreeLocalAdapterRegistry

from ZODB import DB
from ZODB.DemoStorage import DemoStorage
import transaction


from BTrees import family64
OOBTree = family64.OO.BTree
OIBTree = family64.OI.BTree

@interface.implementer(IFoo)
class GlobalUtilityImplementingFoo(object):
    pass

class TestBTreeSiteMan(AbstractTestBase):

    def test_rebuild_swap_class(self):
        base_comps = BLSM(None)
        # replace with "broken"
        base_comps.adapters = _LocalAdapterRegistry()
        base_comps.utilities = _LocalAdapterRegistry()

        sub_comps = BLSM(None)
        sub_comps.adapters = _LocalAdapterRegistry()
        sub_comps.utilities = _LocalAdapterRegistry()

        sub_comps.__bases__ = (base_comps,)

        assert_that(sub_comps.adapters.__bases__, is_((base_comps.adapters,)))
        assert_that(sub_comps.utilities.__bases__, is_((base_comps.utilities,)))

        # Either order works here.
        base_comps.rebuild()
        sub_comps.rebuild()

        # Still in place
        assert_that(sub_comps.adapters.__bases__, is_((base_comps.adapters,)))
        assert_that(sub_comps.utilities.__bases__, is_((base_comps.utilities,)))

        # And they changed type
        assert_that(sub_comps.adapters.__bases__[0], is_(BTreeLocalAdapterRegistry))
        assert_that(sub_comps.utilities.__bases__[0], is_(BTreeLocalAdapterRegistry))
        assert_that(sub_comps.adapters, is_(BTreeLocalAdapterRegistry))
        assert_that(sub_comps.utilities, is_(BTreeLocalAdapterRegistry))


    def _store_base_subs_in_zodb(self, storage):
        from zope.testing.loggingsupport import InstalledHandler
        from contextlib import contextmanager

        log = InstalledHandler('nti.site.site')
        self.addCleanup(log.uninstall)

        @contextmanager
        def fresh_connection():
            db = DB(storage)
            conn = db.open()
            try:
                yield conn
                transaction.commit()
            finally:
                transaction.abort()
                conn.close()
                # Always close the DB to clear connection object caches.
                db.close()

        base_comps = BLSM(None)
        assert isinstance(base_comps, Persistent)
        base_comps.btree_threshold = 0
        base_comps.__name__ = u'base'
        # replace with "broken"
        base_comps.adapters = _LocalAdapterRegistry()
        base_comps.utilities = _LocalAdapterRegistry()

        sub_comps = BLSM(None)
        sub_comps.__name__ = u'sub'
        base_comps.adapters = _LocalAdapterRegistry()
        base_comps.utilities = _LocalAdapterRegistry()
        sub_comps.__bases__ = (base_comps,)

        def check_subs(sub_comps, base_comps, expected_base_kind=BTreeLocalAdapterRegistry):
            assert_that(sub_comps.adapters.__bases__, is_((base_comps.adapters,)))
            assert_that(sub_comps.utilities.__bases__, is_((base_comps.utilities,)))
            assert_that(sub_comps.utilities.__bases__[0], is_(expected_base_kind))

        check_subs(sub_comps, base_comps, _LocalAdapterRegistry)

        with fresh_connection() as conn:
            conn.root()['base'] = base_comps
            conn.root()['sub'] = sub_comps

        # Mimic what the old BLSM.__setstate__ did and
        # convert the data structures.
        with fresh_connection() as conn:
            root = conn.root()
            sub_comps = root['sub']
            base_comps = root['base']

            base_comps._p_activate()
            logs = str(log)
            log.uninstall()
            self.assertIn('ERROR', logs)
            self.assertIn('The LocalSiteManager', logs)
            self.assertIn('has a sub-object', logs)

            # Either order works here
            base_comps.rebuild()
            sub_comps.rebuild()
            check_subs(sub_comps, base_comps)

            assert_that(base_comps.adapters, is_(BTreeLocalAdapterRegistry))
            assert_that(base_comps.utilities, is_(BTreeLocalAdapterRegistry))

        with fresh_connection() as conn:
            root = conn.root()
            sub_comps = root['sub']
            base_comps = root['base']
            check_subs(sub_comps, base_comps)

            assert_that(base_comps.adapters, is_(BTreeLocalAdapterRegistry))
            assert_that(base_comps.utilities, is_(BTreeLocalAdapterRegistry))
            assert_that(sub_comps.adapters, is_(BTreeLocalAdapterRegistry))
            assert_that(sub_comps.utilities, is_(BTreeLocalAdapterRegistry))

    def test_pickle_setstate_swap_class_zodb(self):
        storage = DemoStorage()

        self._store_base_subs_in_zodb(storage)

        db = DB(storage)
        conn = db.open()
        new_base = conn.root()['base']
        new_base._p_activate()
        new_sub = conn.root()['sub']

        # Still in place
        assert_that(new_sub.adapters.__bases__, is_((new_base.adapters,)))
        assert_that(new_sub.utilities.__bases__, is_((new_base.utilities,)))

        # And they changed type
        assert_that(new_sub.adapters.__bases__[0], is_(BTreeLocalAdapterRegistry))
        assert_that(new_sub.utilities.__bases__[0], is_(BTreeLocalAdapterRegistry))

        # But p_changed wasn't set.
        assert_that(new_sub.adapters.__bases__[0],
                    has_property('_p_changed', is_false()))

        transaction.commit()
        conn.close()
        db.close()

        db = DB(storage)
        conn = db.open()
        new_sub = conn.root()['sub']
        # Now, we didn't rewrite the reference so the type stayed the same
        assert_that(type(new_sub.adapters.__bases__[0]), _LocalAdapterRegistry)
        assert_that(new_sub.adapters.__bases__[0],
                    has_property('_p_changed', is_false()))

        # Loading the base changes the types
        new_base = conn.root()['base']
        assert_that(new_base.adapters, is_(BTreeLocalAdapterRegistry))
        assert_that(new_base.adapters,
                    same_instance(new_sub.adapters.__bases__[0]))
        assert_that(new_sub.adapters.__bases__[0], is_(BTreeLocalAdapterRegistry))

        assert_that(new_sub.adapters.__bases__[0],
                    has_property('_p_changed', is_false()))

    def test_pickle_zodb_lookup_adapter(self):
        # Now, we can register a couple adapters in the base, save everything,
        # and look it up in the sub (when the classes don't match)
        storage = DemoStorage()
        self._store_base_subs_in_zodb(storage)

        db = DB(storage)
        conn = db.open()
        new_base = conn.root()['base']
        new_base._p_activate()
        new_sub = conn.root()['sub']

        # Note: this used-to cause btree-ing the map to fail. The
        # implementedBy callable previously had default comparison and can't be
        # stored in a btree. As of zope.interface 4.3.0, this is fixed.
        new_base.registerAdapter(_foo_factory,
                                 required=(object,),
                                 provided=IFoo)

        new_base.registerAdapter(_foo_factory2,
                                 required=(IFoo,),
                                 provided=IMock)

        assert_that(new_base._adapter_registrations, is_(OOBTree))
        assert_that(new_base._adapter_registrations.keys(),
                    contains(
                        ((IFoo,), IMock, u''),
                        ((implementedBy(object),), IFoo, u''),
                    ))
        assert_that(new_base.adapters._provided, is_(OIBTree))
        assert_that(new_base.adapters._adapters[0], is_(OOBTree))

        assert_that(new_base.adapters._adapters[1][IFoo], is_(OOBTree))


        new_base.registerAdapter(_foo_factory2,
                                 required=(IFoo,),
                                 provided=IFoo)

        assert_that(new_base.adapters._adapters[1][IFoo], is_(OOBTree))

        transaction.commit()
        conn.close()
        db.close()

        db = DB(storage)
        conn = db.open()
        new_sub = conn.root()['sub']

        x = new_sub.queryAdapter(RootFoo(), IMock)
        assert_that(x, is_(2))

    def test_register_implemented_by_lookup_utility(self):
        storage = DemoStorage()
        self._store_base_subs_in_zodb(storage)

        db = DB(storage)
        conn = db.open()
        new_base = conn.root()['base']
        new_base._p_activate()
        new_sub = conn.root()['sub']

        new_base.registerUtility(MockSite(),
                                 provided=IFoo)
        provided1 = new_base.adapters._provided
        # In the past, we couldn't register by implemented, but now we can.
        new_base.registerUtility(MockSite(),
                                 provided=implementedBy(MockSite),
                                 name=u'foo')

        provided2 = new_base.adapters._provided
        # Make sure that it only converted once
        assert_that(provided1, is_(same_instance(provided2)))
        assert_that(new_base._utility_registrations, is_(OOBTree))

        assert_that(new_base._utility_registrations.keys(),
                    contains(
                        (IFoo, u''),
                        ((implementedBy(MockSite), u'foo')),
                    ))
        assert_that(new_base.utilities._provided, is_(OIBTree))
        assert_that(new_base.utilities._adapters[0], is_(OOBTree))

        assert_that(new_base.utilities._adapters[0][IFoo], is_(OOBTree))


        transaction.commit()
        conn.close()
        db.close()

        db = DB(storage)
        conn = db.open()
        new_sub = conn.root()['sub']

        x = new_sub.queryUtility(IFoo)
        assert_that(x, is_(MockSite))

        # But it can't actually be looked up, regardless of whether we
        # convert to btrees or not
        x = new_sub.queryUtility(MockSite, u'foo')
        assert_that(x, is_(none()))

    def test_pickle_zodb_lookup_utility(self):
        # Now, we can register a couple utilities in the base, save everything,
        # and look it up in the sub (when the classes don't match)
        storage = DemoStorage()
        self._store_base_subs_in_zodb(storage)

        db = DB(storage)
        conn = db.open()
        new_base = conn.root()['base']
        new_base._p_activate()
        new_sub = conn.root()['sub']

        new_base.registerUtility(MockSite(),
                                 provided=IFoo)
        provided1 = new_base.adapters._provided
        # Previously this would fail. Now it works.
        new_base.registerUtility(MockSite(),
                                 provided=implementedBy(object),
                                 name=u'foo')

        new_base.registerUtility(MockSite(),
                                 provided=IMock,
                                 name=u'foo')

        provided2 = new_base.adapters._provided
        # Make sure that it only converted once
        assert_that(provided1, is_(same_instance(provided2)))
        assert_that(new_base._utility_registrations, is_(OOBTree))

        assert_that(new_base._utility_registrations.keys(),
                    contains(
                        (IFoo, u''),
                        (IMock, u'foo'),
                        (implementedBy(object), u'foo'),
                    ))
        assert_that(new_base.utilities._provided, is_(OIBTree))
        assert_that(new_base.utilities._adapters[0], is_(OOBTree))

        assert_that(new_base.utilities._adapters[0][IFoo], is_(OOBTree))


        transaction.commit()
        conn.close()
        db.close()

        db = DB(storage)
        conn = db.open()
        new_sub = conn.root()['sub']

        x = new_sub.queryUtility(IFoo)
        assert_that(x, is_(MockSite))

        x = new_sub.queryUtility(IMock, u'foo')
        assert_that(x, is_(MockSite))


    def test_convert_with_utility_registered_on_class(self):
        comps = BLSM(None)

        comps.registerUtility(component=MockSite(),
                              provided=implementedBy(object))

        regs = list(comps.registeredUtilities())
        assert_that(regs, has_length(1))

        # But note that we can't actually look this up, regardless of whether
        # we use BTrees or not
        assert_that(regs[0], has_property('provided', implementedBy(object)))

        assert_that(calling(comps.getUtility).with_args(implementedBy(object)),
                    raises(ComponentLookupError))

    def test_convert_with_utility_no_provided(self):
        comps = BLSM(None)

        class AUtility(object):
            # Doesn't implement any interfaces
            pass

        # You can't easily register them this way anyway
        assert_that(calling(comps.registerUtility).with_args(AUtility()),
                    raises(TypeError, "The utility doesn't provide a single interface"))

    def test_convert_with_utility_dynamic_provided(self):
        comps = BLSM(None)
        comps.btree_threshold = 0
        class AUtility(object):
            # Doesn't implement any interfaces
            pass

        autility = AUtility()
        interface.alsoProvides(autility, IFoo)
        comps.registerUtility(autility)
        assert_that(comps._utility_registrations, is_(OOBTree))
        assert_that(comps._utility_registrations.keys(),
                    contains(
                        ((IFoo, u'')),
                    ))
        assert_that(comps.utilities._provided, is_(OIBTree))
        assert_that(comps.utilities._adapters[0], is_(OOBTree))

        assert_that(comps.utilities._adapters[0][IFoo], is_(OOBTree))

    def test_convert_with_adapter_registered_on_class(self):
        comps = BLSM(None)

        comps.registerAdapter(
            _foo_factory,
            required=(object, type('str')),
            provided=IFoo)

        x = comps.getMultiAdapter((object(), 'str'), IFoo)
        assert_that(x, is_(1))

    def test_convert_mark_self_changed(self):
        comps = BLSM(None)

        comps.registerAdapter(
            _foo_factory,
            required=(object, type('str')),
            provided=IFoo)

        comps.registerAdapter(
            _foo_factory2,
            required=(object, type(0)),
            provided=IFoo)

        x = comps.getMultiAdapter((object(), 'str'), IFoo)
        assert_that(x, is_(1))


def _foo_factory(*_args):
    return 1
def _foo_factory2(*_args):
    return 2

class TestPermissiveOOBTree(unittest.TestCase):

    def test_default(self):
        # This is now legacy after BTrees 4.3.2, kept around for
        # ensuring it still works.
        from nti.site.site import _PermissiveOOBTree
        tree = _PermissiveOOBTree()
        key = object() # default comparison

        assert_that(tree.get(key), is_(none()))


from zope.interface.tests.test_adapter import CustomTypesBaseAdapterRegistryTests

class BTreeLocalAdapterRegistryCustomTypesTest(CustomTypesBaseAdapterRegistryTests):

    # Py3: assertRaisesRegexp is deprecated. Make the replacement available
    # everywhere.
    assertRaisesRegex = getattr(unittest.TestCase,
                                'assertRaisesRegex',
                                unittest.TestCase.assertRaisesRegexp)

    def _getMappingType(self):
        return OOBTree

    def _getProvidedType(self):
        return OIBTree

    def _getMutableListType(self):
        from persistent.list import PersistentList
        return PersistentList

    def _getLeafSequenceType(self):
        return self._getMutableListType()

    def _getBaseAdapterRegistry(self):
        return BTreeLocalAdapterRegistry

    def __bt_to_dict(self, bt):
        if not isinstance(bt, (OOBTree, OIBTree)):
            return bt

        return {
            k: self.__bt_to_dict(v)
            for k, v in bt.items()
        }

    def assertEqual(self, first, second, msg=None):
        # BTrees don't support equality tests like
        # dict and PersistentMapping do, so convert them to dicts.
        # They are also found at the top level of the
        # persistent maps, so those need converted too.
        from persistent.list import PersistentList
        if isinstance(first, (OOBTree, OIBTree)) and isinstance(second, (OOBTree, OIBTree)):
            self.assertEqual(type(first), type(second))
            first = self.__bt_to_dict(first)
            second = self.__bt_to_dict(second)
        elif isinstance(first, PersistentList) and isinstance(second, PersistentList):
            self.assertEqual(type(first), type(second))
            first = [self.__bt_to_dict(x) for x in first]
            second = [self.__bt_to_dict(x) for x in second]


        super(BTreeLocalAdapterRegistryCustomTypesTest, self).assertEqual(first, second, msg=msg)

    def test__addValueToLeaf_existing_is_tuple_TypeError(self):
        registry = self._makeOne()
        with self.assertRaisesRegex(TypeError, "Forbidding mutation"):
            registry._addValueToLeaf(('a',), 'b')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Implementations of sites and helpers for working with sites.

"""
# NOTE: unicode_literals is NOT imported!!
from __future__ import print_function, absolute_import, division

__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from BTrees import family64

from zope import component
from zope import interface

from zope.component.hooks import getSite

from zope.site.site import LocalSiteManager
from zope.site.site import _LocalAdapterRegistry

from zope.interface.interfaces import IComponents

from persistent import Persistent

from nti.schema.fieldproperty import createDirectFieldProperties

from nti.schema.schema import SchemaConfigured

from nti.site.interfaces import ISiteMapping
from nti.site.interfaces import SiteNotFoundError

from nti.site.transient import TrivialSite
from nti.site.transient import HostSiteManager


from zope.component.persistentregistry import PersistentComponents

def get_alternate_site_name(site_name):
    """
    Check for a configured ISiteMapping
    """
    site_mapping = component.queryUtility(ISiteMapping, name=site_name)
    if site_mapping is not None:
        return site_mapping.target_site_name
    return None


def find_site_components(site_names, check_alternate=False):
    """
    Return an (global, registered) :class:`.IComponents` implementation named
    for the first virtual site found in the sequence of *site_names*.
    If no such components can be found, returns none.
    """
    for site_name in site_names:
        if not site_name:  # Empty/default. We want the global. This should only ever be at the end
            return None

        if check_alternate:
            site_name = get_alternate_site_name(site_name)
            if site_name is None:
                continue

        components = component.queryUtility(IComponents, name=site_name)
        if components is not None:
            return components
    return None

_find_site_components = find_site_components  # BWC


def get_site_for_site_names(site_names, site=None):
    """
    Return an :class:`.ISite` implementation named for the first virtual site
    found in the sequence of site_names.

    First, we'll attempt to find the registered persistent site; either given
    by the site name or redirected by a registered :class:`ISiteMapping`
    pointing to a persistent site. Otherwise, we'll look for a site without the
    :class:`ISiteMapping` lookup.

    We'll then look a registered persistent site having the same name as the
    registered global components found for *site_names*, then that site will be
    used. Otherwise, if there is only a registered global components, a
    non-persistent site that incorporates those components in the lookup order
    while still incorporating the current (or provided) site will be returned.

    If no such site or components can be found, returns the fallback
    site (the current or provided *site*).

    :param site_names: Sequence of strings giving the virtual host names
        to use.
    :keyword site: If given, this will be the fallback site (and site manager). If
        not given, then the currently installed site will be used.

    .. versionchanged:: 1.2.0
        Look for a :class:`ISiteMapping` registration to map a
        non-persistent site to a persistent site.
    .. versionchanged:: 1.3.0
        Prioritize :class:`ISiteMapping` so that persistent sites can be mapped
        to other persistent sites.
    """

    if site is None:
        site = getSite()

    # assert site.getSiteManager().__bases__ == (component.getGlobalSiteManager(),)
    # Can we find a named site to use?
    site_components = None
    if site_names:
        # First look for an ISiteMapping
        site_components = find_site_components(site_names, check_alternate=True)
        if not site_components:
            site_components = find_site_components(site_names)
    if site_components:
        # Yes we can.
        site_name = site_components.__name__
        # Do we have a persistent site installed in the database? If yes,
        # we want to use that.
        try:
            pers_site = site[u'++etc++hostsites'][site_name]
            site = pers_site
        except (KeyError, TypeError):
            # No, nothing persistent, dummy one up.
            # Note that this code path is deprecated now and not
            # expected to be hit.

            # The site components are only a
            # partial configuration and are not persistent, so we need
            # to use two bases to make it work (order matters) (for
            # example, the main site is almost always the
            # 'nti.dataserver' site, where the persistent intid
            # utilities live; the named sites do not have those and
            # cannot have the persistent nti.dataserver as their real
            # base, so the two must be mixed). They are also not
            # traversable.

            # Host comps used to be simple, but now they may be hierarchacl
            # assert site_components.__bases__ == (component.getGlobalSiteManager(),)
            # gsm = site_components.__bases__[0]
            # assert site_components.adapters.__bases__ == (gsm.adapters,)

            # But the current site, when given, must always be the main
            # dataserver site
            assert isinstance(site, Persistent)
            assert isinstance(site.getSiteManager(), Persistent)

            main_site = site
            # XXX: This easily produces resolution orders that are
            # inconsistent with C3. See test_site.test_no_persistent_site.
            site_manager = HostSiteManager(main_site.__parent__,
                                           main_site.__name__,
                                           site_components,
                                           main_site.getSiteManager())
            site = TrivialSite(site_manager)
            site.__parent__ = main_site
            site.__name__ = site_name

    return site

def get_component_hierarchy(site=None):
    site = getSite() if site is None else site
    # XXX: This is tightly coupled. Note that we assume that the parent
    # site is a container for the persistent sites.
    # There should never be a good reason to need to know this.
    hostsites = site.__parent__
    site_names = (site.__name__,)
    # XXX: Why is this not the same thing as site.getSiteManager()?
    components = find_site_components(site_names)
    while components is not None:
        try:
            name = components.__name__
            if name in hostsites:
                yield components
                components = components.__parent__
            else:
                break
        except AttributeError:  # pragma: no cover
            break

def get_component_hierarchy_names(site=None, reverse=False):
    # XXX This is tightly coupled and there should almost never
    # be a good reason to know this.
    result = [x.__name__ for x in get_component_hierarchy(site)]
    if reverse:
        result.reverse()
    return result

class WrongRegistrationTypeError(TypeError):
    """
    Raised if an adapter registration is of the wrong type.

    .. versionchanged:: 1.0.1
       This is no longer raised by this package.
    """

# This used to catch type errors on get() for default comparison,
# but as of BTrees 4.3.2 that no longer happens. The name needs to
# stay around for BWC with existing pickles.
_PermissiveOOBTree = family64.OO.BTree


class BTreeLocalAdapterRegistry(_LocalAdapterRegistry):
    """
    A persistent adapter registry that can switch its internal
    data structures to be more persistent friendly when they get large.

    .. caution::
       This registry doesn't support registrations on bare
       classes. This is because the Implements and Provides objects
       returned on bare classes do not support comparison or equality
       and hence cannot be used in BTrees. (They only hash and compare
       equal to *themselves*; within the same process this works out
       because of aggressive caching on class objects.) Registering a utility
       to provide a bare class is quite hard to do, in any case. Registering
       adapters to require bare classes is easier but generally not a best practice.

    .. versionchanged:: 3.0.0
       No longer converts any data structures as part of mutating this object.
       Instead, uses the support from zope.interface 5.3 and zope.component 5.0
       to specify the data types to use as they are created on demand.

       Existing persistent registries *must* have the ``rebuild()`` method called
       on them as part of a migration. The best way to do that would be through
       the ``rebuild()`` method on their containing :class:`BTreeLocalSiteManager`.
    """
    # Inherit from _LocalAdapterRegistry for maximum compatibility...we are
    # going to swizzle out classes. Also, it makes sure we are ILocation.

    # Interestingly, we are totally fine to switch out the type from dict
    # to BTree. Much of the actual lookup code is implemented in C, but it calls
    # into Python for _uncached_lookup, which stays in pure python.

    #: The family for the provided map. Defaults to 64-bit maps. I.e., long.
    btree_family = family64

    # Override types from PersistentAdapterRegistry
    _providedType = btree_family.OI.BTree
    _mappingType = btree_family.OO.BTree

    def _addValueToLeaf(self, existing_leaf_sequence, new_item):
        if isinstance(existing_leaf_sequence, tuple):
            # We're mutating unmigrated data. This could lead to data loss
            # if we have a situation from previous versions of this class like
            # BTree -> dict -> tuple; that's about to become
            # BTree -> dict -> PersistentList.
            # Mutations of either the dict or PersistentList won't notify the
            # BTree that it needs to persist itself.
            # In the past, the solution to this was to set the BTree conversion threshold
            # to 0 so that the intermediate dict got converted to a BTree, but that's
            # not possible anymore. So just don't allow it.
            raise TypeError("Forbidding mutation of unmigrated data in %r. Call rebuild()."
                            % self)
        return super(BTreeLocalAdapterRegistry, self)._addValueToLeaf(existing_leaf_sequence,
                                                                      new_item)


class BTreePersistentComponents(PersistentComponents):
    """
    Persistent components that will be friendly to ZODB when they get large.

    Note that despite the name, this class is not Persistent, only its
    internal components are.

    .. caution:: This registry doesn't support bare class registrations.
       See :class:`BTreeLocalAdapterRegistry` for details.
    """

    btree_family = family64

    #: The size at which we will switch from maps to BTrees for registered adapters
    #: and registered utilities (individually). This defaults to the maximum size
    #: of a BTree bucket before it splits. Thus, when we do this, we will wind up with at
    #: least two persistent objects.
    btree_threshold = 30

    def _init_registries(self):
        # NOTE: We cannot simply replace these two attributes at runtime
        # or even in a migration (for example, to upgrade from one type to another type)
        # and expect it to work. If we are the base of some other Components
        # or SiteManager, then these attributes have been copied into the __bases__
        # of *its* adapters and utilities. If we swap out our ivar, then the bases
        # will be out of sync and lookup will be broken. (BTreeLocalSiteManager
        # supposedly keeps track of its subs and so it *could* swap out all of them too.)
        self.adapters = BTreeLocalAdapterRegistry()
        self.utilities = BTreeLocalAdapterRegistry()
        self.adapters.__parent__ = self.utilities.__parent__ = self
        self.adapters.__name__ = u'adapters'
        self.utilities.__name__ = u'utilities'

    def _check_and_btree_map(self, mapping_name):
        # The registrations are mappings that look like this:
        #
        #   {(iface, name): (utility, '', None)}
        btree_type = self.btree_family.OO.BTree
        mapping = getattr(self, mapping_name)
        if not isinstance(mapping, btree_type) and len(mapping) > self.btree_threshold:
            mapping = btree_type(mapping)
            setattr(self, mapping_name, mapping)
            # NOTE: This class is *NOT* Persistent, but its subclass BTreeLocalSiteManager
            # *is*. That's why __setstate__ is there and not here...it doesn't make much sense here.

    def registerUtility(self, *args, **kwargs):  # pylint:disable=arguments-differ
        result = super(BTreePersistentComponents, self).registerUtility(*args, **kwargs)
        self._check_and_btree_map('_utility_registrations')
        return result

    def registerAdapter(self, *args, **kwargs): # pylint:disable=arguments-differ
        result = super(BTreePersistentComponents, self).registerAdapter(*args, **kwargs)
        self._check_and_btree_map('_adapter_registrations')
        return result


class BTreeLocalSiteManager(BTreePersistentComponents, LocalSiteManager):
    """
    Persistent local site manager that will be friendly to ZODB when they
    get large.

    .. caution:: This registry doesn't support bare class registrations.
       See :class:`BTreeLocalAdapterRegistry` for details.

    .. versionchanged:: 3.0.0
       No longer attempts to change the class of the ``adapters`` and ``utilities``
       objects when reading old pickles.

       Instead, you must call this object's ``rebuild()`` method as part of a migration.
       This method will call ``rebuild()`` on the ``adapters`` and ``utilities`` objects,
       and also reset the ``__bases__`` of this object (to its current bases). The
       order (leaves first or roots first) shouldn't matter, as long as all registries
       in an inheritance hierarchy are committed in a single transaction.

       If we detect old versions of the class that haven't been migrated,
       we log an error.
    """
    # pylint:disable=too-many-ancestors

    def __setstate__(self, state):
        super(BTreeLocalSiteManager, self).__setstate__(state)
        for reg in self.adapters, self.utilities:
            if (not isinstance(reg, BTreeLocalAdapterRegistry)
                    and isinstance(reg, _LocalAdapterRegistry)):
                logger.error(
                    "The LocalSiteManager %r has a sub-object %r that is not yet migrated.",
                    self, reg
                )

    def rebuild(self):
        for reg in self.adapters, self.utilities:
            if (not isinstance(reg, BTreeLocalAdapterRegistry)
                    and isinstance(reg, _LocalAdapterRegistry)):
                reg.__class__ = BTreeLocalAdapterRegistry
            reg.rebuild()
        # Setting our bases will cause new references to our *base's*
        # .adapters and .utilities to be saved in the ZODB. As long as they migrate
        # at the same time, they will get written with their new '__class__', even
        # if they are migrated after us.
        self.__bases__ = self.__bases__


@interface.implementer(ISiteMapping)
class SiteMapping(SchemaConfigured):
    """
    Maps one site to another.

    :raises a :class:`SiteNotFoundError` object if no site found
    """
    target_site_name = None
    createDirectFieldProperties(ISiteMapping)

    def get_target_site(self):
        """
        Returns the target site as defined by this mapping.
        """
        current_site = getSite()
        site_names = (self.target_site_name,)
        result = get_site_for_site_names(site_names, site=current_site)
        if result is current_site:
            # Invalid mapping
            raise SiteNotFoundError("No site found for %s" % self.target_site_name)
        return result


# Legacy notes:
# Opening the connection registered it with the transaction manager as an ISynchronizer.
# Ultimately this results in newTransaction being called on the connection object
# at `transaction.begin` time, which in turn syncs the storage. However,
# when multi-databases are used, the other connections DO NOT get this called on them
# if they are implicitly loaded during the course of object traversal or even explicitly
# loaded by name turing an active transaction. This can lead to extra read conflict errors
# (particularly with RelStorage which explicitly polls for invalidations at sync time).
# (Once a multi-db connection has been used, then the next time it would be sync'd. A multi-db
# connection is associated with the same connection to another database for its lifetime, and
# when open()'d will sync all other such connections. Corrollary: ALWAYS go through
# a connection object to get access to multi databases; never go through the database object itself.)

# As a workaround, we iterate across all the databases and sync them manually; this increases the
# cost of handling transactions for things that do not use the other connections, but ensures
# we stay nicely in sync.

# JAM: 2012-09-03: With the database resharding, evaluating the need for this.
# Disabling it.
# for db_name, db in conn.db().databases.items():
#   __traceback_info__ = i, db_name, db, func
#   if db is None: # For compatibility with databases we no longer use
#       continue
#   c2 = conn.get_connection(db_name)
#   if c2 is conn:
#       continue
#   c2.newTransaction()

# Now fire 'newTransaction' to the ISynchronizers, including the root connection
# This may result in some redundant fires to sub-connections.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
.. $Id$
"""

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from zope import interface

from zope.component.zcml import utility

from zope.schema import TextLine

from nti.site.interfaces import ISiteMapping

from nti.site.site import SiteMapping


class ISiteMappingDirective(interface.Interface):
    """
    Register an :class:`ISiteMapping`
    """
    source_site_name = TextLine(title=u"The source site name")

    target_site_name = TextLine(title=u"The target site name")


def registerSiteMapping(_context, source_site_name, target_site_name):
    """
    Create and register a site mapping, as a utility under the `source_site_name`.
    """
    site_mapping = SiteMapping(source_site_name=source_site_name,
                               target_site_name=target_site_name)
    utility(_context, provides=ISiteMapping,
            component=site_mapping, name=source_site_name)



import codecs
from setuptools import setup, find_packages


TESTS_REQUIRE = [
    'fudge',
    'nti.testing >= 3.0.0',
    'pyhamcrest',
    'z3c.baseregistry',
    'zope.testrunner',
    'coverage',
    'zope.testing',
]


def _read(fname):
    with codecs.open(fname, encoding='utf-8') as f:
        return f.read()


setup(
    name='nti.site',
    version=_read('version.txt').strip(),
    author='Jason Madden',
    author_email='jason@nextthought.com',
    description="Opinionated ZODB persistent site implementations",
    long_description=_read('README.rst'),
    url="https://github.com/NextThought/nti.site",
    license='Apache',
    keywords='Site management',
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    zip_safe=True,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    namespace_packages=['nti'],
    tests_require=TESTS_REQUIRE,
    install_requires=[
        'BTrees >= 4.3.2',  # permissive get()
        # test dependencies have this at >= 5.6.0; for consistency,
        # do the same in regular deps.
        'ZODB >= 5.6.0',
        'nti.schema',
        'nti.transactions >= 4.0.0', # nti.transactions.loop
        'persistent',
        'setuptools',
        'six',
        'transaction >= 2.4.0',  # for looser text/byte handling
        'zope.component >= 5.0.0',
        'zope.container',
        'zope.interface >= 5.3.0a1', # Customizable AdapterRegistry data
        'zope.location',
        'zope.proxy',
        'zope.site >= 4.4.0', # Proper site cleanup.
        'zope.traversing',
    ],
    extras_require={
        'test': TESTS_REQUIRE,
        'docs': [
            'Sphinx',
            'repoze.sphinx.autointerface',
            'sphinx_rtd_theme',
        ] + TESTS_REQUIRE # To be able to import nti.site.testing
    },
)



# -*- coding: utf-8 -*-
"""
Tests for nti.site.tests.testing.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

from nti.site import testing

class TestPrintTree(unittest.TestCase):

    def setUp(self):
        super(TestPrintTree, self).setUp()
        testing.setHooks()

    def tearDown(self):
        testing.resetHooks()
        super(TestPrintTree, self).tearDown()

    def test_print_tree(self):

        class Trans(testing.persistent_site_trans):
            printed = None
            def on_application_and_sites_installed(self, folder):
                super(Trans, self).on_application_and_sites_installed(folder)
                folder._p_jar.root()['key'] = 'value'
                self.printed = testing.format_tree(folder._p_jar.root())


        # open for the side-effects
        t = Trans(testing.default_db_factory())
        with t:
            pass

        printed = t.printed

        self.assertIn('<Connection Root Dictionary>', printed)
        self.assertIn('<ISite,IRootFolder>: Application', printed)
        self.assertIn('<ISite,IMainApplicationFolder>: dataserver2', printed)
        self.assertIn('dataserver2 -> dataserver2', printed)
        self.assertIn('nti.dataserver -> dataserver2', printed)
        self.assertIn('nti.dataserver_root -> Application', printed)
        self.assertIn('len=', printed)

    def test_print_basic(self):
        printed = testing.format_tree('a string')
        self.assertIn('a string', printed)

    def test_print_no_len(self):
        printed = testing.format_tree(object())
        self.assertNotIn('len=', printed)

    def test_print_unknown_type(self):
        called = []
        def show_unknown(o):
            called.append(42)
            return str(o)

        printed = testing.format_tree(object(), show_unknown=show_unknown)
        self.assertIn('object', printed)
        self.assertEqual(called, [42])

class TestCurrentDBSiteTrans(testing.SiteTestCase):

    def test_find_site(self):
        with testing.persistent_site_trans(site_name='foobar'):
            pass

class TestIndependent(unittest.TestCase):

    @testing.uses_independent_db_site
    def test_sets_db_no_layer(self):
        self.assertTrue(hasattr(self, 'db'))

    @testing.uses_independent_db_site(installer_kwargs={'site_name': None})
    def test_keyword_argument(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division

__docformat__ = "restructuredtext en"

from nti.testing import zodb

from .. import testing


current_db_site_trans = testing.persistent_site_trans # BWC, remove in 2021
mock_db_trans = current_db_site_trans # BWC, remove in 2021
reset_db_caches = zodb.reset_db_caches # BWC, remove in 2021
WithMockDS = testing.uses_independent_db_site # BWC, remove in 2021
SharedConfiguringTestLayer = testing.SharedConfiguringTestLayer # BWC, remove in 2021
SiteTestCase = testing.SiteTestCase # BWC, remove in 2021



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Helpers for running jobs in specific sites.

"""

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

import warnings


from zope import component
from zope import interface

from zope.component.hooks import site as current_site

from ZODB.interfaces import IDatabase

from nti.transactions.loop import TransactionLoop

from nti.site.interfaces import SiteNotInstalledError

from nti.site.interfaces import ITransactionSiteNames
from nti.site.interfaces import ISiteTransactionRunner

from nti.site.site import get_site_for_site_names

logger = __import__('logging').getLogger(__name__)


# transaction >= 2 < 2.1.1 needs text; Transaction 1 wants
# bytes (generally). Transaction 2.1.1 and above will work with either,
# but text is preferred.
def _tx_string(s):
    return s.decode('utf-8', 'replace') if isinstance(s, bytes) else s


class _RunJobInSite(TransactionLoop):

    _connection = None

    def __init__(self, *args, **kwargs):
        self.site_names = kwargs.pop('site_names')
        self.job_name = kwargs.pop('job_name')
        self.side_effect_free = kwargs.pop('side_effect_free')
        self.root_folder_name = kwargs.pop('root_folder_name')
        super(_RunJobInSite, self).__init__(*args, **kwargs)

    def describe_transaction(self, *args, **kwargs):
        if self.job_name:
            return _tx_string(self.job_name)
        # Derive from the function
        func = self.handler
        name = getattr(func, '__name__', '')
        doc = getattr(func, '__doc__', '')
        if name == '_': # "Anonymous" function; transaction convention
            name = ''
        note = None
        if doc:
            note = ((name + '\n\n') if name else '') + doc
        elif name:
            note = name

        note = _tx_string(note) if note else None

        return note

    def run_handler(self, *args, **kwargs): # pylint:disable=arguments-differ
        sitemanc = self._connection.root()[self.root_folder_name]
        # Put into a policy if need be
        sitemanc = get_site_for_site_names(self.site_names, sitemanc)

        with current_site(sitemanc):
            if component.getSiteManager() != sitemanc.getSiteManager():
                raise SiteNotInstalledError("Hooks not installed?")
            return self.handler(*args, **kwargs)

    def setUp(self):
        # After the transaction manager has been put into explicit
        # mode, open the connection. This lets it perform certain
        # optimizations.
        db = component.getUtility(IDatabase)
        self._connection = db.open()

    def tearDown(self):
        if self._connection is not None:
            try:
                self._connection.close()
            finally:
                self._connection = None


_marker = object()

def get_possible_site_names(*args, **kwargs):
    """
    Helper to find the most applicable site names.

    This uses the :class:`.ITransactionSiteNames` utility.
    """
    utility = component.queryUtility(ITransactionSiteNames)
    result = utility(*args, **kwargs) if utility is not None else None
    return result

@interface.provider(ISiteTransactionRunner)
def run_job_in_site(func,
                    retries=0,
                    sleep=None,
                    site_names=_marker,
                    job_name=None,
                    side_effect_free=False,
                    root_folder_name=u'nti.dataserver'):
    """
    Runs the function given in `func` in a transaction and dataserver local
    site manager. See :class:`.ISiteTransactionRunner`

    :return: The value returned by the first successful invocation of `func`.
    """

    # site_names is deprecated, we want to start preserving
    # the current site. Because the current site should be based on the
    # current site names FOR NOW, preserving the current site names
    # is equivalent. THIS IS CHANGING though.
    if site_names is not _marker:
        warnings.warn("site_names is deprecated. "
                      "Call this already in the appropriate site",
                      FutureWarning)
    else:
        site_names = get_possible_site_names()

    return _RunJobInSite(
        func,
        retries=retries,
        sleep=sleep,
        site_names=site_names,
        job_name=job_name,
        side_effect_free=side_effect_free,
        root_folder_name=root_folder_name
    )()

run_job_in_site.__doc__ = ISiteTransactionRunner['__call__'].getDoc()
